%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%                                %%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%      INPAINTING MAIN FILE      %%%%%%%%%%%%%%%%%%%%%%%     
%%%%%%%%%%%%%%%%%%%%                                %%%%%%%%%%%%%%%%%%%%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% comparison of SPAIN variants:
%     A-SPAIN
%     S-SPAIN OMP
%     S-SPAIN H

close all;
clear variables;
clc;

%% input file settings   
fprintf('Loading audio: ')

% available samples:
%     acoustic_guitar
%     double_bass
%     flute
%     organ
%     piano
%     PianoVoice_clean
%     sample
%     string_quartet
%     violin
%     vivaldi_storm

sample = 'violin';

[data, fs] = audioread(strcat('samples/',sample,'.wav'));
param.Ls = length(data);

fprintf(strcat(sample,'.wav\n'));


%% gap parameters
gap_length = 800; % gap length in samples
gap_start = 2; % gap start in s


%% settings
fprintf('Setting up the frame parameters\n')

% frame parameters
red = 4; % redundancy of DFT in each block
param.w = 1200; % window length
param.a = 300; % window shift
param.wtype = 'hann';
param.M = param.w;
param.offset = 'full'; % 'half' / 'full' / 'none'

% The offset parameter ensures that the processed signal blocks are
% distributed symmetrically with respect to the center of the gap.
% It is done by moving the starting point of the transform in time domain.
% Offset 'full' corresponds to the situation where middle of the gap
% overlaps with middle index of a block. Offset 'half', on the other hand,
% has the middle of the gap on the axis of two adjacent blocks.

% DFT parameters
param.F = frame('dft');
param.F.redundancy = red; % non-native, our own additional parameter
param.F.frana = @(insig)dft([insig; zeros(length(insig)*(param.F.redundancy-1),1)]);
param.F.frsyn = @(insig)postpad(idft(insig),length(insig)/param.F.redundancy);

% paramsolver parameters
paramsolver.s = 1; % increment of k
paramsolver.r = 1; % every r-th iteration increment k by s   
paramsolver.epsilon = 0.1;  % stopping criterion of termination function
paramsolver.maxit = ceil(floor(param.w*param.F.redundancy/2+1)*paramsolver.r/paramsolver.s); % maximum number of iterations
paramsolver.store_snr = 1; 
paramsolver.store_obj = 1;


%% degrading the signal 
fprintf('Generating degraded signal\n\n')
fprintf('   Gap start:   %3.2f s\n',gap_start);
fprintf('   Gap length: %4.2f ms (%d samples)\n',1000*gap_length/fs,gap_length);
s = round(fs*gap_start);
h = gap_length;
f = s + h - 1;
param.mask = true(param.Ls,1);
param.mask(s:f) = false;
data_gapped = gapping(data, param.mask); % degrading the original signal


%% shortening the degraded signal
switch param.offset
    case 'full'
        c = ceil((s+f)/2);
        k = floor((c-1)/param.a);
        d = 1 + k*param.a ;
        offset = c-d;
    case 'half'
        c = ceil((s+f)/2);
        k = floor((c-1)/param.a);
        d = 1 + k*param.a + ceil(param.a/2);
        offset = c-d;
    otherwise
        offset = 0;
end
[q,Q,~,~,~,~,~,~,~,~,param.Ls] = min_sig_supp_2(param.w,param.a,param.M,s,f,param.Ls,1,offset);
mask = param.mask; % for SNR evaluation
param.mask = param.mask(q:Q); % for SPAIN algorithm, where we work with shortened signal
% inicialization of restored signal...
data_rec_1 = data_gapped; % ... for A-SPAIN
data_rec_2 = data_gapped; % ... for S-SPAIN OMP
data_rec_3 = data_gapped; % ... for S-SPAIN H
data_gapped = data_gapped(q:Q); % shortening gapped data for the algorithm in the same way as with the mask
fprintf('\n-----------------------------------------------\n');

%% A-SPAIN
param.algorithm = 'aspain';
fprintf('\nStarting the A-SPAIN algorithm\n\n');

t = tic; 
[data_rec_1(q:Q), snr_iter_1, obj_iter_1] = spain_segmentation(data_gapped, param, paramsolver, data(q:Q));
time = toc(t);

% time
fprintf('   Result obtained in %4.3f seconds.\n', time);
% SNR in the gap
snr_rec = snr_n(data(~mask), data_rec_1(~mask));
fprintf('   SNR of the reconstructed signal is %4.3f dB.\n', snr_rec);
fprintf('\n-----------------------------------------------\n');


%% S-SPAIN OMP
param.algorithm = 'sspain';
paramsolver.f_update = 'OMP';
fprintf('\nStarting the S-SPAIN OMP algorithm\n(this may take up to a few minutes)\n\n');

t = tic; 
[data_rec_2(q:Q), snr_iter_2, obj_iter_2] = spain_segmentation(data_gapped, param, paramsolver, data(q:Q));
time = toc(t);

% time
fprintf('   Result obtained in %4.3f seconds.\n', time);
% SNR in the gap
snr_rec = snr_n(data(~mask), data_rec_2(~mask));
fprintf('   SNR of the reconstructed signal is %4.3f dB.\n', snr_rec);
fprintf('\n-----------------------------------------------\n');


%% S-SPAIN H
param.algorithm = 'sspain';
paramsolver.f_update = 'H';
fprintf('\nStarting the S-SPAIN H algorithm\n\n');

t = tic; 
[data_rec_3(q:Q), snr_iter_3, obj_iter_3] = spain_segmentation(data_gapped, param, paramsolver, data(q:Q));
time = toc(t);

% time
fprintf('   Result obtained in %4.3f seconds.\n', time);
% SNR in the gap
snr_rec = snr_n(data(~mask), data_rec_3(~mask));
fprintf('   SNR of the reconstructed signal is %4.3f dB.\n', snr_rec);
fprintf('\n-----------------------------------------------\n');


%% plot of original signal and reconstruction
figure();
t = (s-100:f+100)/fs;
plot(t,data(s-100:f+100)); hold on
plot(t,data_rec_1(s-100:f+100)); hold on;
plot(t,data_rec_2(s-100:f+100)); hold on;
plot(t,data_rec_3(s-100:f+100)); hold on;
title({'Original signal and reconstruction using SPAIN'; '(plotting the filled gap and surroundings)'});
legend('original','A-SPAIN','S-SPAIN OMP','S-SPAIN H');
xlim([t(1) t(end)]);
xlabel('time [s]');


%% plot objective functions
% randomly choose one of the processed blocks
[~,n] = size(obj_iter_1);
indexes = 1:n;
processed_blocks = indexes(~isnan(obj_iter_1(1,:)));
block = processed_blocks(randi(length(processed_blocks)));

% plot
figure;
plot(obj_iter_1(:,block));
hold on;
plot(obj_iter_2(:,block));
hold on;
plot(obj_iter_3(:,block));
legend('A-SPAIN','S-SPAIN OMP','S-SPAIN H');
title(sprintf('Norm of the residuum in randomly chosen block (number %d)',block));
xlabel('Iteration number');
ylabel('Norm of the residuum');


%% plot spectrograms
% frame setting for the spectrogram
number = 120;
margin = 1000;
F_plot = frametight(frame('dgtreal',{param.wtype,param.w},param.a,param.M,'timeinv'));

% spectrograms as subplots
figure();
subplot(2,2,1);
colormap jet
plotframe(F_plot,frana(F_plot,data),fs,number);
xlim([s-margin,f+margin]/fs);
title('Original signal');

subplot(2,2,2);
colormap jet
plotframe(F_plot,frana(F_plot,data_rec_1),fs,number);
xlim([s-margin,f+margin]/fs);
title('A-SPAIN');

subplot(2,2,3);
colormap jet
plotframe(F_plot,frana(F_plot,data_rec_2),fs,number);
xlim([s-margin,f+margin]/fs);
title('S-SPAIN OMP');

subplot(2,2,4);
colormap jet
plotframe(F_plot,frana(F_plot,data_rec_2),fs,number);
xlim([s-margin,f+margin]/fs);
title('S-SPAIN H');